@@ -57,6 +57,17 @@ OGR_TYPE_DEF = {
 }
 
 
+OGR_GEOM_DEF = {
+                1: 'point',
+                2: 'line',
+                3: 'polygon',
+                4: 'multipoint',
+                5: 'multilinestring',
+                6: 'multipolygon',
+                0: 'geometry',
+                100: 'no geometry'}
+
+
 class Raster(object):
     """
     Class to read and write rasters from/to files and numpy arrays
@@ -720,25 +731,31 @@ class Raster(object):
     def get_coords(xy_list,
                    pixel_size,
                    tie_point,
+                   order_as_yx=False,
                    pixel_center=True):
         """
         Method to convert pixel locations to image coords
         :param xy_list: List of tuples [(x1,y1), (x2,y2)....]
         :param pixel_size: tuple of x and y pixel size. The signs of the pixel sizes (+/-) are as in GeoTransform
         :param tie_point: tuple of x an y coordinates of tie point for the xy list
+        :param order_as_yx: If the order of coordinates is [(y1,x1),..] instead of [(x1,y1),..]
         :param pixel_center: If the center of the pixels should be returned instead of the top corners (default: True)
         :return: List of coordinates in tie point coordinate system
         """
+        pixel_xy_list = xy_list.copy()
 
-        if type(xy_list) != list:
-            xy_list = [xy_list]
+        if type(pixel_xy_list) != list:
+            pixel_xy_list = [pixel_xy_list]
+
+        if order_as_yx:
+            pixel_xy_list = list([y, x] for x, y in pixel_xy_list)
 
         add_const = (float(pixel_size[0])/2.0) * pixel_center, \
                     (float(pixel_size[1])/2.0) * pixel_center
 
         return list((float(xy[0]) * float(pixel_size[0]) + tie_point[0] + add_const[0],
                      float(xy[1]) * float(pixel_size[1]) + tie_point[1] + add_const[1])
-                    for xy in xy_list)
+                    for xy in pixel_xy_list)
 
     @staticmethod
     def get_locations(coords_list,
@@ -1284,22 +1301,24 @@ class Raster(object):
                         pixel_center: bool (True) default
                                      Used only if the pass_pixel_coords flag is set to true.
                                      Returns the pixel centers of each pixel if set as true
-                                     else returns top left corners
-                        min_pixels: Minimum number of pixels for reducer to calculate a value.
+                                     else returns top left corners if set as false
+                        min_pixels: Minimum number of pixels for reducer to calculate a value. (default: 1)
                                     If the number of pixels are less than this, then an empty list is returned
                         replace: If the pixel values in Raster.array object should be replaced with
-                                 the reduced value? This will only be computed of min_pixels condition is
+                                 the reduced value? This will only be computed if min_pixels condition is
                                  satisfied and will change the Raster array. Use with caution.
                         reducer: Valid keywords: 'mean','median','max',
                                                  'min', 'percentile_xx' where xx is percentile from 1-99
 
-        :return: Dictionary of dictionaries
-                {internal_id: {'values': [[band1, ], ], 'coordinates': [(x1, y1), ]}, }
-                if geom_id is supplied then internal_id is the supplied geom_id
+        :return: list of dictionaries
+                [ {'values': [[band1, ], ], 'coordinates': [(x1, y1), ]}, 'zyx_loc': [[band, y, x], ], ]
+                The order is the same as the supplied geometry wkt string list
 
         todo: Implement separation of overlapping feature extractions
 
         """
+        max_burn_chunks = 65534  # gdal cannot burn more than this number to a layer (rasterization)
+
         if band_order is None:
             band_order = list(range(self.shape[0]))
 
@@ -1333,13 +1352,18 @@ class Raster(object):
         if 'min_pixels' in kwargs:
             min_pixels = kwargs['min_pixels']
         else:
-            min_pixels = None
+            min_pixels = 1
 
         if 'replace' in kwargs:
             replace = kwargs['replace']
         else:
             replace = False
 
+        if 'verbose' in kwargs:
+            verbose = kwargs['verbose']
+        else:
+            verbose = False
+
         # define band order
         if band_order is None:
             band_order = list(range(0, self.shape[0]))
@@ -1348,6 +1372,9 @@ class Raster(object):
         if not self.init or self.array is None:
             self.initialize()
 
+        if verbose:
+            Opt.cprint(self.__repr__())
+
         # make wkt strings into a list
         if type(wkt_strings) not in (list, tuple):
             wkt_strings = [wkt_strings]
@@ -1361,15 +1388,23 @@ class Raster(object):
         # make internal tiles
         self.make_tile_grid(*tile_size)
 
+        if verbose:
+            Opt.cprint('Processing {} tiles: '.format(str(self.ntiles)))
+
         # prepare dict struct
-        out_geom_extract = dict()
-        for internal_id, _ in id_geom_list:
-            out_geom_extract[internal_id] = {'values': [], 'coordinates': [], 'xyz_loc': []}
+        out_geom_extract = {internal_id: {'values': [], 'coordinates': [], 'zyx_loc': []}
+                            for internal_id, _ in id_geom_list}
 
         geom_id_order = list(internal_id for internal_id, _ in id_geom_list)
 
         # list of sample ids
-        for tile in self.tile_grid:
+        for tile_indx, tile in enumerate(self.tile_grid):
+
+            if verbose:
+                Opt.cprint('Processing tile {} of {} with dimensions {}x{}'.format(str(tile_indx + 1),
+                                                                                   str(self.ntiles),
+                                                                                   str(tile['block_coords'][3]),
+                                                                                   str(tile['block_coords'][2])), ' ')
 
             # create tile geometry from bounds
             tile_geom = ogr.CreateGeometryFromWkt('POLYGON(({}))'.format(', '.join(list(' '.join([str(x), str(y)])
@@ -1380,16 +1415,26 @@ class Raster(object):
 
             # check if the geometry intersects and
             # place all same geometry types together
+            geom_counter = 0
             geom_by_type = {}
             for samp_id, samp_geom in id_geom_list:
                 if tile_geom.Intersects(samp_geom):
+                    geom_counter += 1
                     geom_type = samp_geom.GetGeometryType()
                     if geom_type not in geom_by_type:
                         geom_by_type[geom_type] = [(samp_id, samp_geom)]
                     else:
                         geom_by_type[geom_type].append((samp_id, samp_geom))
 
-            # check is any geoms are available
+            if verbose:
+                Opt.cprint('extracting {} geometries ...'.format(str(geom_counter)), ' ')
+                prnt_str_list = []
+                for k, v in geom_by_type.items():
+                    prnt_str_list.append('{} : {}'.format(str(OGR_GEOM_DEF[k]),
+                                                          str(len(v))))
+                Opt.cprint(', '.join(prnt_str_list))
+
+            # check if any geoms are available
             if len(geom_by_type) > 0:
                 for geom_type, geom_list in geom_by_type.items():
 
@@ -1397,97 +1442,115 @@ class Raster(object):
                     _, _, rows, cols = tile['block_coords']
                     tie_pt_x, tie_pt_y = tile['tie_point']
 
-                    # create tile empty raster in memory
-                    target_ds = gdal.GetDriverByName('MEM').Create('tmp',
-                                                                   cols,
-                                                                   rows,
-                                                                   1,
-                                                                   gdal.GDT_UInt16)
+                    # divide into chunks if more than max_burn_chunks
+                    if len(geom_list) > max_burn_chunks:
+                        n_chunks = int(np.ceil(len(geom_list) / max_burn_chunks))
+                        chunks = list(geom_list[(chunk_indx * max_burn_chunks): ((chunk_indx+1) * max_burn_chunks)]
+                                      if (chunk_indx < (n_chunks-1)) else geom_list[(chunk_indx * max_burn_chunks):]
+                                      for chunk_indx in range(n_chunks))
+                    else:
+                        chunks = [geom_list]
 
-                    # set pixel size and tie point
-                    target_ds.SetGeoTransform((tie_pt_x,
-                                               self.transform[1],
-                                               0,
-                                               tie_pt_y,
-                                               0,
-                                               self.transform[5]))
+                    for chunk in chunks:
+                        chunk_geom_list = list([indx, elem[1]] for indx, elem in enumerate(chunk))
+                        chunk_indx_geom_indx_relation = {indx: elem[0] for indx, elem in enumerate(chunk)}
 
-                    # set raster projection
-                    target_ds.SetProjection(self.crs_string)
+                        # create tile empty raster in memory
+                        target_ds = gdal.GetDriverByName('MEM').Create('tmp',
+                                                                       cols,
+                                                                       rows,
+                                                                       1,
+                                                                       gdal.GDT_UInt16)
 
-                    # create vector in memory
-                    burn_driver = ogr.GetDriverByName('Memory')
-                    burn_datasource = burn_driver.CreateDataSource('mem_source')
-                    burn_spref = osr.SpatialReference()
-                    burn_spref.ImportFromWkt(self.crs_string)
-                    burn_layer = burn_datasource.CreateLayer('tmp_lyr',
-                                                             srs=burn_spref,
-                                                             geom_type=geom_type)
+                        # set pixel size and tie point
+                        target_ds.SetGeoTransform((tie_pt_x,
+                                                   self.transform[1],
+                                                   0,
+                                                   tie_pt_y,
+                                                   0,
+                                                   self.transform[5]))
 
-                    # attributes
-                    fielddefn = ogr.FieldDefn('fid', ogr.OFTInteger)
-                    result = burn_layer.CreateField(fielddefn)
-                    layerdef = burn_layer.GetLayerDefn()
+                        # set raster projection
+                        target_ds.SetProjection(self.crs_string)
 
-                    geom_burn_val = 1
-                    geom_dict = {}
-                    for geom_id, geom in geom_list:
-                        # create features in layer
-                        temp_feature = ogr.Feature(layerdef)
-                        temp_feature.SetGeometry(geom)
-                        temp_feature.SetField('fid', geom_burn_val)
-                        burn_layer.CreateFeature(temp_feature)
-                        geom_dict[geom_burn_val] = geom_id
-                        geom_burn_val += 1
+                        # create vector in memory
+                        burn_driver = ogr.GetDriverByName('Memory')
+                        burn_datasource = burn_driver.CreateDataSource('mem_source')
+                        burn_spref = osr.SpatialReference()
+                        burn_spref.ImportFromWkt(self.crs_string)
+                        burn_layer = burn_datasource.CreateLayer('tmp_lyr',
+                                                                 srs=burn_spref,
+                                                                 geom_type=geom_type)
 
-                    gdal.RasterizeLayer(target_ds,
-                                        [1],
-                                        burn_layer,
-                                        None,  # transformer
-                                        None,  # transform
-                                        [1],
-                                        ['ALL_TOUCHED=TRUE',
-                                         'ATTRIBUTE=FID'])
+                        # attributes
+                        primary_field = ogr.FieldDefn('fid', ogr.OFTInteger)
+                        result = burn_layer.CreateField(primary_field)
+                        layerdef = burn_layer.GetLayerDefn()
 
-                    # read mask band as array
-                    temp_band = target_ds.GetRasterBand(1)
-                    mask_arr = temp_band.ReadAsArray()
+                        geom_burn_val = 1
+                        geom_dict = {}  # dict of burn val: geom id
+                        for geom_id, geom in chunk_geom_list:
+                            # create features in layer
+                            temp_feature = ogr.Feature(layerdef)
+                            temp_feature.SetGeometry(geom)
+                            temp_feature.SetField('fid', geom_burn_val)
+                            burn_layer.CreateFeature(temp_feature)
+                            geom_dict[geom_burn_val] = geom_id
+                            geom_burn_val += 1
 
-                    min_burn_val = min(list(geom_dict.keys()))
-                    max_burn_val = max(list(geom_dict.keys()))
+                        gdal.RasterizeLayer(target_ds,
+                                            [1],
+                                            burn_layer,
+                                            None,  # transformer
+                                            None,  # transform
+                                            [0],
+                                            ['ALL_TOUCHED=TRUE',
+                                             'ATTRIBUTE=FID'])
 
-                    pixel_xy_loc = np.where((mask_arr >= min_burn_val) & (mask_arr <= max_burn_val))
-                    burn_vals = mask_arr[pixel_xy_loc]
-                    pixel_xy_loc_list = np.array(list(zip(*pixel_xy_loc)))
+                        # read mask band as array
+                        temp_band = target_ds.GetRasterBand(1)
+                        mask_arr = temp_band.ReadAsArray()
 
-                    burn_dict = {}
-                    for geom_burn_val in range(min_burn_val, max_burn_val + 1):
-                        burn_dict[geom_burn_val] = pixel_xy_loc_list[np.where(burn_vals == geom_burn_val)].tolist()
+                        min_burn_val = min(list(geom_dict.keys()))
+                        max_burn_val = max(list(geom_dict.keys()))
 
-                    for geom_burn_val, geom_id in geom_dict.items():
-                        # make list of unmasked pixels
-                        if pass_pixel_coords:
-                            # get coordinates
-                            out_geom_extract[geom_id]['coordinates'] += self.get_coords(burn_dict[geom_burn_val],
-                                                                                        (self.transform[1],
-                                                                                         self.transform[5]),
-                                                                                        tile['tie_point'],
-                                                                                        pixel_center)
+                        pixel_yx_loc = np.where((mask_arr >= min_burn_val) & (mask_arr <= max_burn_val))
+                        burn_vals = mask_arr[pixel_yx_loc]
+                        pixel_yx_loc_arr = np.array(list(zip(*pixel_yx_loc)))
 
-                        # get band values from tile array
-                        out_geom_extract[geom_id]['values'] += list(tile_arr[band_order, x, y].tolist()
-                                                                    for x, y in burn_dict[geom_burn_val])
+                        burn_dict = {burn_val: pixel_yx_loc_arr[np.where(burn_vals == burn_val)].tolist()
+                                     for burn_val in list(set(burn_vals))}
 
-                        out_geom_extract[geom_id]['xyz_loc'] += list([[band_index] + list(xy_loc)]
-                                                                     for xy_loc in burn_dict[geom_burn_val]
-                                                                     for band_index in band_order)
+                        for geom_burn_val, geom_id in geom_dict.items():
+                            # convert chunk geom id to actual/global geom id
+                            actual_geom_id = chunk_indx_geom_indx_relation[geom_id]
+
+                            if geom_burn_val in burn_dict:
+
+                                if pass_pixel_coords:
+                                    # get coordinates
+                                    out_geom_extract[actual_geom_id]['coordinates'] += \
+                                        self.get_coords(burn_dict[geom_burn_val],
+                                                        (self.transform[1], self.transform[5]),
+                                                        tile['tie_point'],
+                                                        pixel_center=pixel_center,
+                                                        order_as_yx=True)
+                                # get band values from tile array
+                                out_geom_extract[actual_geom_id]['values'] += list(tile_arr[band_order, y, x].tolist()
+                                                                                   for y, x in
+                                                                                   burn_dict[geom_burn_val])
+
+                                out_geom_extract[actual_geom_id]['zyx_loc'] += list([[band_index] + list(yx_loc)]
+                                                                                    for yx_loc in
+                                                                                    burn_dict[geom_burn_val]
+                                                                                    for band_index in band_order)
 
             if reducer is not None:
                 warned = False
                 for geom_id, geom_dict in out_geom_extract.items():
                     try:
                         n_pixels = len(geom_dict['values'])
-                        if min_pixels is None:
+                        if min_pixels is None or (not (0 < min_pixels <= n_pixels)):
                             min_pixels = n_pixels
 
                         if n_pixels >= min_pixels:
@@ -1496,9 +1559,9 @@ class Raster(object):
                                                                  axis=0).tolist()
 
                             if replace:
-                                pixel_loc = (np.array(geom_dict['xyz_loc'])[:, 0],
-                                             np.array(geom_dict['xyz_loc'])[:, 1],
-                                             np.array(geom_dict['xyz_loc'])[:, 2])
+                                pixel_loc = (np.array(geom_dict['zyx_loc'])[:, 0],
+                                             np.array(geom_dict['zyx_loc'])[:, 1],
+                                             np.array(geom_dict['zyx_loc'])[:, 2])
 
                                 self.array[pixel_loc] = geom_dict['values']
 
@@ -1766,6 +1829,10 @@ class Raster(object):
         else:
             vrt_ds = None
 
+    def close(self):
+        """Close the raster file handle from GDAL"""
+        self.datasource = None
+
 
 class MultiRaster:
     """
