@@ -542,26 +542,31 @@ class Vector(object):
 
     @staticmethod
     def get_osgeo_geom(geom_string,
-                       geom_type='wkt'):
+                       geom_type='wkt',
+                       buffer=None):
         """
         Method to return a osgeo geometry object
         :param geom_string: Wkt or json string
         :param geom_type: 'wkt', 'json', or 'wkb
+        :param buffer: Amount in geometry coordinates to buffer the geometry
         :return: osgeo geometry object
         """
         if geom_type == 'wkt':
             try:
-                return ogr.CreateGeometryFromWkt(geom_string)
+                return ogr.CreateGeometryFromWkt(geom_string) if buffer is None else \
+                    ogr.CreateGeometryFromWkt(geom_string).Buffer(buffer)
             except:
                 return
         elif geom_type == 'json':
             try:
-                return ogr.CreateGeometryFromJson(geom_string)
+                return ogr.CreateGeometryFromJson(geom_string) if buffer is None else \
+                    ogr.CreateGeometryFromJson(geom_string).Buffer(buffer)
             except:
                 return
         elif geom_type == 'wkb':
             try:
-                return ogr.CreateGeometryFromWkb(geom_string)
+                return ogr.CreateGeometryFromWkb(geom_string) if buffer is None else \
+                    ogr.CreateGeometryFromWkb(geom_string).Buffer(buffer)
             except:
                 return
         else:
